using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Cryptography;
using System.Security.Cryptography.Xml;
using System.Security.Cryptography.X509Certificates;
using System.Xml;
using CAPICOM;
using System.IO;
using System.Text.RegularExpressions;

namespace ConexionPlatino
{
    public class InformacionFirmaPLatino
    {
        public string fichero;
        public string signature;
        public string nif;
        public string nombre;
        public string cargo;
        public string signaturevalue;
        public bool bvalida;
        public InformacionFirmaPLatino()
        {
            fichero = signature = nif = nombre = cargo = signaturevalue = "";
            bvalida = false;
        }
    }
    public class UtilidadesPLatino
    {
        /// <summary>
        /// Firmamos fichero binario con XMLdSig Enveloped de su hash
        /// </summary>
        /// <param name="URIString"></param>
        /// <param name="certificate"></param>
        /// <returns></returns>
        public XmlElement FirmarFicheroEnveloped(string URIString, X509Certificate2 certificate, string[] filtroIssuer)
        {

            // Cargar el certificado X509.
            X509Certificate2 MSCert;
            if (certificate == null)
            {
                CAPICOM.Certificate Cert = SeleccionarCertificado(filtroIssuer);

                if (Cert == null) return null;

                X509Store store = new X509Store(StoreName.My, StoreLocation.CurrentUser);
                store.Open(OpenFlags.ReadOnly);
                X509Certificate2Collection listaCert = store.Certificates.Find(X509FindType.FindBySerialNumber, Cert.SerialNumber.ToLower(), false);
                MSCert = listaCert[0];
            }
            else
                MSCert = certificate;

            //Creamos un XML con el Hash del documento
            XmlDocument document = new XmlDocument();
            XmlNode node = document.CreateNode(XmlNodeType.Element, "", "HashDocumento", null);
            node.InnerText = CalculateHashFromFile(URIString);
            document.AppendChild(node);

            // Crear un objeto SignedXml.
            SignedXml signedXml = new SignedXml();

            // Asignar la clave para el objeto SignedXml.
            signedXml.SigningKey = MSCert.PrivateKey;

            // Creamos el objeto a firmar
            DataObject dataObject = new DataObject();
            dataObject.Data = document.ChildNodes;
            string nid = Guid.NewGuid().ToString();
            string id = "Documento" + nid;
            dataObject.Id = id;

            // Aadimos el objeto
            signedXml.AddObject(dataObject);

            // Crear una referencia a la firma.
            Reference reference = new Reference();

            // Aadir el URI transferido al objeto de referencia.
            reference.Uri = "#" + id;

            // Aadir una transformacin si el URI es un archivo XML.
            if (URIString.EndsWith("xml"))
            {
                reference.AddTransform(new XmlDsigC14NTransform());
            }

            // Aadir la referencia al objeto SignedXml.
            signedXml.AddReference(reference);

            // Crear un objeto KeyInfo nuevo.
            KeyInfo keyInfo = new KeyInfo();
            // Cargar el certificado en un objeto KeyInfoX509Data 
            // y aadir al objeto KeyInfo.
            KeyInfoX509Data keyInfoData = new KeyInfoX509Data(MSCert);
            keyInfoData.AddIssuerSerial(MSCert.Subject, MSCert.GetSerialNumberString());
            keyInfo.AddClause(keyInfoData);

            RSA key = RSA.Create();
            keyInfo.AddClause(new RSAKeyValue(key));


            signedXml.KeyInfo = keyInfo;

            // Calcular la firma.
            signedXml.ComputeSignature();

            string s = signedXml.ToString();

            // Obtener la representacin XML de la firma y guardar 
            // a un objeto XmlElement.           
            XmlElement xmlDigitalSignature = signedXml.GetXml();


            // Sacamos la firma como string
            return xmlDigitalSignature;
        }

        /// <summary>
        /// Valida la firma XMLdSig Enveloped de un hash de un fichero binario
        /// </summary>
        /// <param name="firma"></param>
        /// <param name="fileName"></param>
        /// <param name="certificate"></param>
        /// <returns></returns>
        public Boolean ValidarFirmaFicheroEnveloped(string firma, string fileName, X509Certificate2 certificate)
        {
            byte[] bytes = File.ReadAllBytes(fileName);
            return ValidarFirmaEnveloped(firma, bytes, certificate);
        }

        /// <summary>
        /// Valida la firma XMLdSig Enveloped de array de bytes
        /// </summary>
        /// <param name="firma"></param>
        /// <param name="bytesFichero"></param>
        /// <param name="certificate"></param>
        /// <returns></returns>
        public Boolean ValidarFirmaEnveloped(string firma, byte[] bytesFichero, X509Certificate2 certificate)
        {
            bool valid = false;            

            try
            {
                XmlDocument doc = new XmlDocument();
                doc.LoadXml(firma);
                XmlNodeList nodeList;
        

                //Comprobamos los hash de los ficheros
                string hash = CalculateHash(bytesFichero);

                nodeList = doc.GetElementsByTagName("Signature");
                if (nodeList == null) return false;                               

                XmlElement elem = null;
                foreach (XmlElement HashObj in nodeList[0].ChildNodes)
                {
                    try
                    {
                        elem = HashObj["HashDocumento"];
                        if (elem != null)
                            break;
                    }
                    catch { }
                }

                if (elem == null || !hash.Equals(elem.InnerXml)) return false;

                SignedXml signedXml = new SignedXml();
                signedXml.LoadXml((XmlElement)nodeList[0]);

                //Si se especifica un certificado validamos la firma contra ese certificado
                valid = (certificate != null) ? signedXml.CheckSignature(certificate, true) : signedXml.CheckSignature();
                
            }
            catch { }

            return valid;
        }

        /// <summary>
        /// Firma una cadena en PCKS7 con un X509Certificate
        /// </summary>
        /// <param name="msg"></param>
        /// <param name="ClientCert"></param>
        /// <param name="bDetached"></param>
        /// <returns></returns>
        public string SignMsg(string msg, X509Certificate2 ClientCert, bool bDetached)
        {
            CAPICOM.Store store = new CAPICOM.Store();
            store.Open(CAPICOM.CAPICOM_STORE_LOCATION.CAPICOM_CURRENT_USER_STORE, "My", CAPICOM.CAPICOM_STORE_OPEN_MODE.CAPICOM_STORE_OPEN_READ_ONLY);

            CAPICOM.Certificate certificate = null;
            foreach (Certificate cert in store.Certificates)
            {
                if (cert.SerialNumber.Equals(ClientCert.SerialNumber))
                {
                    certificate = cert;
                    break;
                }
            }

            if (certificate == null) return null;

            return (SignMsg(msg, certificate, bDetached));
        }

        /// <summary>
        /// Calcula MD5 de una cadena
        /// </summary>
        /// <param name="sTextToHash"></param>        
        /// <returns></returns>
        public static string MD5HashString(string sTextToHash)
        {
            MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider();
            byte[] byValue;
            byte[] byHash;

            byValue = System.Text.Encoding.UTF8.GetBytes(sTextToHash);
            byHash = md5.ComputeHash(byValue);
            md5.Clear();

            return Convert.ToBase64String(byHash);
        }

        /// <summary>
        /// Calcula HASH de un fichero
        /// </summary>
        /// <param name="filename"></param>        
        /// <returns></returns>
        public static string CalculateHashFromFile(string filename)
        {
            // hash generation from file

            // read file bytes
            byte[] bytes = File.ReadAllBytes(filename);
            return CalculateHash(bytes);


        }

        /// <summary>
        /// Calcula HASH de un array de bytes
        /// </summary>
        /// <param name="bytes"></param>        
        /// <returns></returns>
        public static string CalculateHash(byte[] bytes)
        {
            // md5 calculation
            MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider();
            byte[] output = md5.ComputeHash(bytes);
            string hash = Convert.ToBase64String(output);

            // returning
            return hash;
        }

        /// <summary>
        /// Firma una cadena en PCKS7 con un certificado CAPICOM
        /// </summary>
        /// <param name="msg"></param>
        /// <param name="ClientCert"></param>
        /// <param name="bDetached"></param>
        /// <returns></returns>
        public string SignMsg(string msg, CAPICOM.Certificate ClientCert, bool bDetached)
        {
            return SignMsg(Encoding.UTF8.GetBytes(msg), ClientCert, bDetached);

        }

        /// <summary>
        /// Firma array de buytes en PCKS7 con un certificado CAPICOM
        /// </summary>
        /// <param name="msg"></param>
        /// <param name="ClientCert"></param>
        /// <param name="bDetached"></param>
        /// <returns></returns>
        public string SignMsg(byte[] b, CAPICOM.Certificate ClientCert, bool bDetached)
        {
            CAPICOM.Attribute TimeAttribute = new CAPICOM.AttributeClass();
            CAPICOM.SignedData signedData = new CAPICOM.SignedDataClass();
            CAPICOM.Utilities u = new CAPICOM.UtilitiesClass();
            signedData.set_Content(u.ByteArrayToBinaryString(b));
            CAPICOM.Signer signer = new CAPICOM.Signer();
            signer.Certificate = ClientCert;


            TimeAttribute.Name = CAPICOM.CAPICOM_ATTRIBUTE.CAPICOM_AUTHENTICATED_ATTRIBUTE_SIGNING_TIME;
            TimeAttribute.Value = DateTime.Now.Date;
            signer.AuthenticatedAttributes.Add(TimeAttribute);

            string signature = signedData.Sign(signer, bDetached, CAPICOM.CAPICOM_ENCODING_TYPE.CAPICOM_ENCODE_BASE64);


            return signature;


        }

        /// <summary>
        /// Seleccionar certificado  dentro de los del filtro  
        /// </summary>
        /// <param name="FilterIssuer"></param>
        /// <returns></returns>
        public CAPICOM.Certificate SeleccionarCertificado(string[] FilterIssuer)
        {
            CAPICOM.Store store = new CAPICOM.Store();
            store.Open(CAPICOM.CAPICOM_STORE_LOCATION.CAPICOM_CURRENT_USER_STORE, "My", CAPICOM.CAPICOM_STORE_OPEN_MODE.CAPICOM_STORE_OPEN_READ_ONLY);
            CAPICOM.Certificates posiblesCerts = ((CAPICOM.Certificates)store.Certificates).Find(CAPICOM.CAPICOM_CERTIFICATE_FIND_TYPE.CAPICOM_CERTIFICATE_FIND_TIME_VALID, DateTime.Today, true);
            CAPICOM.Certificates certs = new CAPICOM.Certificates();

            if (FilterIssuer != null && FilterIssuer.Length > 0)
            {
                foreach (CAPICOM.Certificate Cert in posiblesCerts)
                {
                    string compania = Cert.IssuerName;
                    compania = compania.ToUpper();
                    for (short i = 0; i < FilterIssuer.Length; i++)
                    {
                        if (compania.ToUpper().Contains(FilterIssuer[i]))
                        {
                            certs.Add(Cert);
                            break;
                        }
                    }

                }

            }
            else
            {
                foreach (CAPICOM.Certificate Cert in posiblesCerts) certs.Add(Cert);
            }


            CAPICOM.Certificates selectCerts;


            try
            {
                //selectCerts = ((CAPICOM.Certificates)store.Certificates).Select("Lista de Certificados Disponibles", "Seleccione el certificado a utilizar para la firma", false);
                selectCerts = certs.Select("Lista de Certificados Disponibles", "Seleccione el certificado a utilizar para la firma", false);

            }
            catch
            {
                return null;
            }
            if (selectCerts.Count <= 0) return null;

            return (CAPICOM.Certificate)selectCerts[1];
        }
        #region RUTINAS ESPECIFICAS DE UNIFICA
        // Sacamos la firma como string
        public bool CreateFicheroSignature(string firma, XmlElement signature, bool bCreate, bool bReplace)
        {
            XmlDocument doc = new XmlDocument();


            if (bCreate) File.Delete(firma);

            if (!File.Exists(firma))
            {

                XmlDeclaration xmlDeclaration = doc.CreateXmlDeclaration("1.0", "utf-8", null);
                doc.AppendChild(xmlDeclaration);

                XmlNode signatures = doc.CreateNode(XmlNodeType.Element, "Signatures", null);

                signatures.AppendChild(doc.ImportNode(signature, true));
                doc.AppendChild(signatures);
                doc.Save(firma);


                return true;
            }


            doc.Load(firma);
            XmlElement elemSignatures = doc["Signatures"];

            if (bReplace)
            {
                try
                {

                    XmlElement elemSerial = signature["KeyInfo"]["X509Data"]["X509IssuerSerial"]["X509SerialNumber"];


                    foreach (XmlElement elemsig in elemSignatures.ChildNodes)
                    {
                        try
                        {

                            XmlElement elemSerial2 = elemsig["KeyInfo"]["X509Data"]["X509IssuerSerial"]["X509SerialNumber"];

                            if (elemSerial2.InnerText.Equals(elemSerial.InnerText))
                            {
                                elemSignatures.RemoveChild(elemsig);
                                break;
                            }


                        }
                        catch { }

                    }
                }
                catch { }


            }


            elemSignatures.AppendChild(doc.ImportNode(signature, true));
            doc.Save(firma);
            return true;
        }
        public string CreateFicheroSignatureUnifica(string fichero, XmlElement signature)
        {
            XmlDocument doc = new XmlDocument();
            FileInfo file = new FileInfo(fichero);
            
            string[] separador = { "_firmado","." };
            string[] t2 = file.Name.Split(separador, StringSplitOptions.None);            
            int nFirmasInicial = t2.Length - 2;
            int nFirmas;

            if (nFirmasInicial == 0)
            {
                XmlDeclaration xmlDeclaration = doc.CreateXmlDeclaration("1.0", "utf-8", null);
                doc.AppendChild(xmlDeclaration);
                XmlNode signatures = doc.CreateNode(XmlNodeType.Element, "Signatures", null);

                signatures.AppendChild(doc.ImportNode(signature, true));
                doc.AppendChild(signatures);
                nFirmas = 1;
            }

            else
            {
                doc.Load(fichero+"_sign");
                XmlElement elemSignatures = doc["Signatures"];

                try
                {

                    XmlElement elemSerial = signature["KeyInfo"]["X509Data"]["X509IssuerSerial"]["X509SerialNumber"];


                    foreach (XmlElement elemsig in elemSignatures.ChildNodes)
                    {
                        try
                        {

                            XmlElement elemSerial2 = elemsig["KeyInfo"]["X509Data"]["X509IssuerSerial"]["X509SerialNumber"];

                            if (elemSerial2.InnerText.Equals(elemSerial.InnerText))
                            {
                                elemSignatures.RemoveChild(elemsig);
                                break;
                            }


                        }
                        catch { }

                    }
                }
                catch { }
                
                elemSignatures.AppendChild(doc.ImportNode(signature, true));

                XmlNodeList node = elemSignatures.GetElementsByTagName("Signature");
                nFirmas = node.Count;

            }

            string sfilename;
            if (nFirmas == nFirmasInicial)            
                sfilename = fichero;                            
            else
            {
                sfilename = file.DirectoryName + "\\"+t2[0];
                for (short i = 0; i < nFirmas; i++)
                    sfilename += "_firmado";
                sfilename += file.Extension;
                File.Copy(fichero, sfilename, true);
                
                
            }
            File.Delete(sfilename + "_sign");
            doc.Save(sfilename + "_sign");
            return sfilename;
        }
        public bool informacionFirma(string fichero, InformacionFirmaPLatino infFirma, int nFirma)
        {
            XmlDocument doc = new XmlDocument();
            int n = 0;
            string snif = "";
            string cn = "";

            XmlElement elemSignatures;

            if (!File.Exists(fichero) || !File.Exists(fichero + "_sign") || nFirma < 0) return false;

            try
            {
                doc.Load(fichero + "_sign");
                elemSignatures = doc["Signatures"];
                if (elemSignatures == null || elemSignatures.ChildNodes.Count <= nFirma) return false;
            }
            catch 
            {
                return false;
            }

            try
            {
                string nifPattern = "(((s|S)(e|E)(r|R)(i|I)(a|A)(l|L)(n|N)(u|U)(m|M)(b|B)(e|E)(r|R))|(OID.2.5.4.5)|((n|N)(i|I)(f|F)))\\s*=?\\s*(?<nif>[a-zA-Z0-9]{9})";
                string cnPattern = "CN=\"(?<cn>(.*?)\\s*)(\"|$)";



                foreach (XmlElement elemsig in elemSignatures.ChildNodes)
                {
                    if (n < nFirma)
                    {
                        n++;
                        continue;
                    }

                    try
                    {

                        XmlElement elemSerial2 = elemsig["KeyInfo"]["X509Data"]["X509IssuerSerial"]["X509IssuerName"];
                        Match resultado = Regex.Match(elemSerial2.InnerText, nifPattern);

                        if (resultado.Success)
                        {
                            snif = resultado.Groups["nif"].Value;
                            resultado = Regex.Match(elemSerial2.InnerText, cnPattern);
                            cn = "";

                            if (resultado.Success)
                                cn = resultado.Groups["cn"].Value;

                            else
                            {
                                cnPattern = "NOMBRE\\s*(?<cn>(.*?)\\s*)(-|$)";
                                resultado = Regex.Match(elemSerial2.InnerText, cnPattern);
                                if (resultado.Success)
                                    cn = resultado.Groups["cn"].Value;
                                else
                                {
                                    cnPattern = "CN=(?<cn>(.*?)\\s*)(,|$)";
                                    resultado = Regex.Match(elemSerial2.InnerText, cnPattern);
                                    if (resultado.Success)
                                        cn = resultado.Groups["cn"].Value;
                                }
                            }

                            infFirma.fichero = fichero;
                            infFirma.signature = elemsig.OuterXml;
                            infFirma.nif = snif;
                            infFirma.nombre = cn;
                            infFirma.bvalida = ValidarFirmaFicheroEnveloped(elemsig.OuterXml, fichero, null);
                            infFirma.cargo = "";



                            foreach (XmlElement cargoObj in elemsig.GetElementsByTagName("Object"))
                            {
                                XmlElement cargoElem = cargoObj["Cargo"];
                                if (cargoElem != null)
                                {
                                    infFirma.cargo = cargoElem.InnerText;
                                    break;
                                }
                            }
                            XmlElement signaturevalue = elemsig["SignatureValue"];
                            infFirma.signaturevalue = signaturevalue.InnerText;
                            return true;



                        }
                    }
                    catch
                    {
                    }

                }
            }
            catch
            {
            }

            return false;


        }

        public string informacionFirmaTexto(InformacionFirmaPLatino infFirma)        
        {
            return infFirma.cargo + ", NIF: " + infFirma.nif + ", Nombre: " + infFirma.nombre;
        }
        public string certificadoFirma(string firma)
        {
            XmlDocument doc = new XmlDocument();
            try
            {
                doc.LoadXml(firma);
                XmlElement elemCertificado = doc.FirstChild["KeyInfo"]["X509Data"]["X509Certificate"] ;
                if (elemCertificado == null) return null;
                else
                    return elemCertificado.InnerText;
                
            }
            catch
            {
                return null;
            }

            
        }
        public string dameCargoFirma(string signature)
        {
            XmlDocument doc = new XmlDocument();
            try
            {
                doc.LoadXml(signature);
                XmlElement elemsig = doc["Signature"];
                foreach (XmlElement cargoObj in elemsig.GetElementsByTagName("Object"))
                {
                    XmlElement cargoElem = cargoObj["Cargo"];
                    if (cargoElem != null)
                        return cargoElem.InnerText;

                }
                return "";
            }
            catch
            {
                return "";
            }

            
        }
        #endregion 
    }

}
